﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.SceneManagement;
using UnityEngine.UI;

public class Game : MonoBehaviour
{
    public Message MessageSource;
    public InputMessage Input;
    public Message OutgoingMessage;

    public Image SourceX, SourceO;

    public RectTransform GameOverRoot;
    public RectTransform TutorialRoot;

    public Text PlayerScoreDisplay;
    public Text ScoreCounterSource;

    public float GenerationScaler = 1;

    public AudioSource PlayOnReceive, PlayOnSend, PlayOnMatch, PlayOnFail;

    readonly List<Message> IncomingMsgs = new List<Message>();
    readonly bool[] stuckMsgs = new bool[50]; //stuck to the ground;

    CodeGrid latestSent = null;
    float nextMsg = 0;
    int msgCount = 0;

    int playerScore = 0;

    float Speed { get { return 50 + 5 * (msgCount / 10); } }

    bool isClearing = false; // pauses game during send/clear animation
    bool isPullingRows = false;

    public void ReloadScene()
    {
        SceneManager.LoadScene(0);
    }

    // Use this for initialization
    void Awake()
    {
        Input.OnSend += Input_OnSend;
        OutgoingMessage.transform.localScale = Vector3.zero;

        SourceO.gameObject.SetActive(false);
        SourceX.gameObject.SetActive(false);
        GameOverRoot.gameObject.SetActive(false);
        ScoreCounterSource.gameObject.SetActive(false);
    }

    IEnumerator Start()
    {
        yield return null;
        MessageSource.gameObject.SetActive(false);
    }

    private void Input_OnSend(CodeGrid obj)
    {
        PlayOnSend.Play();
        latestSent = obj;
    }

    // Update is called once per frame
    void Update()
    {
        if (isClearing)
            return;

        if (updateIncoming()) // returns true on game over
        {
            StartCoroutine(animateGameOver());
        }
        updateMessages();
        if (latestSent != null)
        {
            OutgoingMessage.SetCodes(latestSent);
            StartCoroutine(animateOutgoing());
            latestSent = null;
        }
    }

    void updateMessages()
    {

        float curTop = Input.Top;
        for (int i = 0; i < IncomingMsgs.Count; i++)
            updateMessagePosition(i, Time.deltaTime);
    }

    bool updateMessagePosition(int index, float deltaTime)
    {
        float curTop = index == 0 ? Input.Top : IncomingMsgs[index - 1].Top;
        IncomingMsgs[index].RectTform.anchoredPosition += Vector2.down * Speed * deltaTime;
        float y = IncomingMsgs[index].RectTform.anchoredPosition.y;
        stuckMsgs[index] = (y < curTop);
        if (stuckMsgs[index])
            IncomingMsgs[index].RectTform.anchoredPosition = new Vector2(IncomingMsgs[index].RectTform.anchoredPosition.x, curTop);
        return stuckMsgs[index];
    }

    bool updateIncoming()
    {
        if ((nextMsg -= Time.deltaTime) <= 0)
        {
            msgCount++;
            nextMsg += UnityEngine.Random.value * 3.6f + (msgCount > 70 ? 1 : msgCount > 15 ? 2f : 3.5f);
            nextMsg *= Mathf.Max(GenerationScaler, 0.01f);
            var newMsg = Instantiate(MessageSource.gameObject) as GameObject;
            newMsg.SetActive(true);
            (newMsg.transform as RectTransform).SetParent(MessageSource.transform.parent, false);
            var msg = newMsg.GetComponent<Message>();
            IncomingMsgs.Add(msg);
            msg.SetCodes(getNextCodes());
            msg.AnimatePopup();
            PlayOnReceive.Play();

            // check for Game Over condition
            float topMsg = IncomingMsgs.Count <= 1 ? 0 : IncomingMsgs[IncomingMsgs.Count - 2].Top;
            if (topMsg > msg.Bottom)
            {
                if (!pushDownMessage(IncomingMsgs.Count - 2, topMsg - msg.Bottom))
                    return true; // GAME OVER!!!!
            }
        }

        return false;
    }

    bool pushDownMessage(int index, float targetDist)
    {
        float startY = IncomingMsgs[index].Bottom;

        bool wasPushedOk = !updateMessagePosition(index, targetDist / Speed);

        if (wasPushedOk) return true; //easy case because plenty of space
        else if (index == 0) return false; // easy case bc no more space below us

        // hard case
        targetDist -= (startY - IncomingMsgs[index].Bottom);
        bool pushedNext = pushDownMessage(index - 1, targetDist);

        if (pushedNext)
            updateMessagePosition(index, targetDist / Speed);

        return pushedNext;
    }

    CodeGrid getNextCodes()
    {
        int colCountMax = Mathf.Clamp((msgCount / 10) + 3, 5, 8);
        int rowCount = 1;// Mathf.Min(3, msgCount / 100) + 1; // cant have multiline incoming until some fixes are made to checking code
        CodeGrid newGrid = new CodeGrid(UnityEngine.Random.Range(3, colCountMax), rowCount);
        int codeCount = CodeImageDB.Inst.Codes.Length;
        int codeMin = msgCount * 3;// UnityEngine.Random.Range(0, codeCount);
        int codeMax = codeMin + (codeCount / (msgCount < 5 ? 4 : msgCount < 25 ? 3 : msgCount < 50 ? 2 : 1));
        for (int i = 0; i < newGrid.Codes.Length; i++) newGrid.Codes[i] = CodeImageDB.Inst.Codes[UnityEngine.Random.Range(codeMin, codeMax) % codeCount];

        int spaceCount = Mathf.Clamp(msgCount / 25, 0, Mathf.Min(10, newGrid.Codes.Length / 3));

        for (int i = 0; i < spaceCount; i++)
            newGrid.Codes[UnityEngine.Random.Range(0, newGrid.Codes.Length)] = 0;

        return newGrid;
    }

    // the send-message anim is kinda complex, so just to keep my head around it all...
    //
    // 1. pause gameplay, pop up the SENT message
    // 2. pull down incoming messages (COMP_SET) that MAY match the input (ie for a 3-line input, pull down 3 lines, if needed)
    // 3. move COMP_SET right to overlap with SENT
    // 4. simultaneously, move SENT offscreen to the right, and COMP_SET to the left, offscreen for success, back to original x-pos for failure
    // 5. continue gameplay
    //

    IEnumerator animateOutgoing()
    {
        // pauses stuff
        isClearing = true;

        int outRows = Mathf.Min(OutgoingMessage.Codes.Rows, IncomingMsgs.Count);

        StartCoroutine(pulldownMessages(outRows)); // we dont yield because we use bool to wait, but let them run simultaneously

        for (float t = 0, dur = 0.45f; t < dur; t += Time.deltaTime)
        {
            OutgoingMessage.transform.localScale = Vector3.one * (t / dur);
            yield return null;
        }
        while (isPullingRows)
            yield return null;

        bool success = checkCanClearRows();

        Vector2 outpos = OutgoingMessage.RectTform.anchoredPosition;
        float messageX = outRows > 0 ? IncomingMsgs[0].RectTform.anchoredPosition.x : 0;
        float targetX = success ? 450 : 0;

        if (outRows > 0) //animate the input and targets into the center
        {
            float width = OutgoingMessage.Codes.Cols * 0.5f * (OutgoingMessage.IconSize + OutgoingMessage.PaddingX + IncomingMsgs[0].IconSize + IncomingMsgs[0].PaddingX);
            float offset = (500 - width) / 2;

            for (float t = 0, dur = 0.25f; t < dur; t += Time.deltaTime)
            {
                OutgoingMessage.RectTform.anchoredPosition = outpos + Vector2.left * offset * (t / dur);

                for (int i = 0; i < outRows; i++)
                    IncomingMsgs[i].RectTform.anchoredPosition = new Vector2(messageX, IncomingMsgs[i].RectTform.anchoredPosition.y) + Vector2.right * offset * (t / dur);

                yield return null;
            }
            OutgoingMessage.RectTform.anchoredPosition = outpos + Vector2.left * offset;
            for (int i = 0; i < outRows; i++)
                IncomingMsgs[i].RectTform.anchoredPosition = new Vector2(messageX, IncomingMsgs[i].RectTform.anchoredPosition.y) + Vector2.right * offset;
            yield return new WaitForSeconds(0.4f);
        }

        if (success)
        {
            StartCoroutine(animateScoreEarned(OutgoingMessage.Codes.Codes.Length * 100, OutgoingMessage.Codes.Rows));
        }

        (success ? PlayOnMatch : PlayOnFail).Play();

        //animate input and targets back outward, offscreen for target if correct
        for (float t = 0, dur = 0.25f; t < dur; t += Time.deltaTime)
        {
            OutgoingMessage.RectTform.anchoredPosition = outpos + Vector2.right * 450 * (t / dur);

            for (int i = 0; i < outRows; i++)
                IncomingMsgs[i].RectTform.anchoredPosition = new Vector2(messageX, IncomingMsgs[i].RectTform.anchoredPosition.y) + Vector2.left * targetX * (t / dur) * (t / dur);

            yield return null;
        }

        OutgoingMessage.transform.localScale = Vector3.zero;
        OutgoingMessage.RectTform.anchoredPosition = outpos;



        if (success)
        {
            for (int i = 0; i < outRows; i++)
                Destroy(IncomingMsgs[i].gameObject);

            IncomingMsgs.RemoveRange(0, outRows);
        }


        isClearing = false;
    }

    IEnumerator pulldownMessages(int rowCount)
    {
        isPullingRows = true;

        for (int i = 0; i < rowCount; i++)
        {
            while (true) // loop until a message hits the bottom
            {
                float y = IncomingMsgs[i].RectTform.anchoredPosition.y;
                updateMessagePosition(i, 0.8f);
                if (IncomingMsgs[i].RectTform.anchoredPosition.y == y)
                    break;
                else
                    yield return null;
            }
        }
        yield return null;

        isPullingRows = false;
    }

    bool checkCanClearRows() // called during send animation to see if the player gave correct input
    {
        CodeGrid outCodes = OutgoingMessage.Codes;

        if (outCodes.Rows > IncomingMsgs.Count)
            return false;

        int cols = 0;
        for (int i = 0; i < outCodes.Rows; i++)
            cols = Mathf.Max(cols, IncomingMsgs[i].Codes.Cols);

        if (outCodes.Cols != cols)
            return false;

        CodeGrid compareGrid = new CodeGrid(cols, outCodes.Rows);
        for (int i = 0; i < compareGrid.Rows; i++)
        {
            int ix = compareGrid.Rows - (i + 1);
            for (int x = 0; x < compareGrid.Cols && x < IncomingMsgs[ix].Codes.Cols; x++)
                compareGrid.Set(x, i, IncomingMsgs[ix].Codes.Get(x, 0));
        }

        for (int i = 0; i < outCodes.Codes.Length; i++)
            if (outCodes.Codes[i] != compareGrid.Codes[i])
                return false;

        return true;
    }

    List<Image> xo = new List<Image>();

    IEnumerator animateGameOver()
    {
        isClearing = true; //reappropriate this to pause the main game stuff

        GameOverRoot.gameObject.SetActive(true);
        for (float t = 0, dur = 0.5f; t < dur; t += Time.deltaTime)
        {
            GameOverRoot.anchoredPosition = new Vector2(GameOverRoot.anchoredPosition.x, -GameOverRoot.sizeDelta.y * (1 - (t / dur) * (t / dur)));
            yield return null;
        }
        GameOverRoot.anchoredPosition = new Vector2(GameOverRoot.anchoredPosition.x, 0);
    }

    Coroutine scoreRoutine = null;

    IEnumerator animateScoreEarned(int baseScore, int mult)
    {
        TutorialRoot.gameObject.SetActive(false);

        ScoreCounterSource.gameObject.SetActive(true);
        ScoreCounterSource.text = "+" + baseScore.ToString("N0") + (mult <= 1 ? "" : "\nx" + mult);
        Color c = ScoreCounterSource.color;
        c.a = 1;
        ScoreCounterSource.color = c;

        for (float t = 0, dur = 0.2f; t < dur; t += Time.deltaTime)
        {
            ScoreCounterSource.transform.localScale = Vector2.one * (t / dur);
            yield return null;
        }

        if (mult > 1)
        {
            yield return new WaitForSeconds(0.3f);
            ScoreCounterSource.text = "+" + (baseScore * mult).ToString("N0");
        }

        playerScore += baseScore * mult;
        PlayerScoreDisplay.text = playerScore.ToString("N0");

        Vector2 pos = ScoreCounterSource.rectTransform.anchoredPosition;
        for (float t = 0, dur = 0.9f; t < dur; t += Time.deltaTime)
        {
            ScoreCounterSource.rectTransform.anchoredPosition += Vector2.up * Time.deltaTime * 300;
            c.a = 1 - (t / dur);
            ScoreCounterSource.color = c;
            yield return null;
        }

        ScoreCounterSource.rectTransform.anchoredPosition = pos;
        ScoreCounterSource.gameObject.SetActive(false);

        scoreRoutine = null;
    }
}
