﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;

public class Message : MonoBehaviour
{
    protected const int MaxCols = 8;

    public Image Background;
    public RectTransform IconsRoot;

    public int IconSize;
    public int PaddingX, PaddingY;

    public bool IsRightSide;
    

    public bool TrimVisibleCols;
    

    public CodeGrid Codes { get; private set; }

    public float Bottom { get { return rectTransform.anchoredPosition.y; } }
    public float Top { get { return rectTransform.anchoredPosition.y + (IconSize + PaddingY) * Codes.Rows; } }

    public RectTransform RectTform { get { return rectTransform; } }

    protected RectTransform rectTransform;

    protected readonly List<Image> images = new List<Image>();

    bool isInit = false;

    public int GetVisibleCols()
    {
        int max = 0;
        for (int y = 0; y < Codes.Rows; y++)
            max = Mathf.Max(max, GetVisibleCols(y));
        return Mathf.Max(1, max);
    }

    public int GetVisibleCols(int row)
    {
        for(int col = Codes.Cols; col > 0; col--)
        {
            if (Codes.Get(col - 1, row) != 0)
                return col;
        }
        return 0;
    }

    public void AnimatePopup()
    {
        StartCoroutine(popupAnim(0.5f));
    }

    public void SetCodes(CodeGrid codes)
    {
        Codes = new CodeGrid(codes);
        Codes.Trim(true, true);
        Codes.OnSizeChanged += onResize;
        if(isInit) onResize();
    }

    protected void init()
    {
        rectTransform = transform as RectTransform;

        images.Clear();
        IconsRoot.GetComponentsInChildren(images);

        if (Codes == null) Codes = new CodeGrid(MaxCols, 1);
        Codes.OnSizeChanged -= onResize;
        Codes.OnSizeChanged += onResize;

        Debug.Log(gameObject.name + ": Bot=" + Bottom + ", Top=" + Top);

        isInit = true;
        onResize();
        Debug.Log(gameObject.name + ": Bot=" + Bottom + ", Top=" + Top);
    }

    protected virtual void onResize()
    {
        int cols = TrimVisibleCols ? GetVisibleCols() : Codes.Cols;
        Background.rectTransform.sizeDelta = new Vector2(cols * (IconSize + PaddingX), Codes.Rows * (IconSize + PaddingY) - 1);
        rebuildIcons(false);
    }

    protected void rebuildIcons(bool useVisibleCols)
    {
        int cols = useVisibleCols ? GetVisibleCols() : Codes.Cols;
        int rows = Codes.Rows;

        if (IsRightSide)
        {
            Debug.Log("right.");
        }

        int count = images.Count;
        while (count > rows * cols)     // delete excess images if needed
        {
            DestroyImmediate(images[count - 1].gameObject);
            images.RemoveAt(count - 1);
            count--;
        }
        while (count < rows * cols)     // add images if needed
        {
            GameObject newImageObj = Instantiate(images[0].gameObject) as GameObject;
            newImageObj.GetComponent<RectTransform>().SetParent(images[0].rectTransform.parent, false);
            images.Add(newImageObj.GetComponent<Image>());
            count++;
        }

        Debug.Log("Rebuilding count=" + count + ", cols=" + cols + ", rows=" + rows);
        Vector2 posOffset = new Vector2(PaddingX, PaddingY) * 0.5f;
        Vector2 rightLeftOffset = !IsRightSide ? Vector2.zero : (Vector2.left * Codes.Cols * (PaddingX + IconSize));
        for (int i = 0, x, y; i < count; i++)   // set positions for our rows*cols images
        {
            x = i % cols;
            y = (rows - 1) - i / cols;
            images[i].rectTransform.sizeDelta = Vector2.one * IconSize;
            images[i].rectTransform.anchoredPosition = posOffset + rightLeftOffset + new Vector2(x * (IconSize + PaddingX), y * (IconSize + PaddingY));
        }
        refreshImages();
    }

    protected void refreshImages()
    {
        for(int i = 0; i < images.Count; i++)
        {
            CodeImageDB.Inst.SetImage(images[i], Codes.Codes[i]);
        }
    }

    IEnumerator popupAnim(float time)
    {
        float totalTime = time;
        transform.localScale = Vector3.zero;
        while(time > 0)
        {
            time -= Time.deltaTime;
            transform.localScale = Vector3.one * (1 - Mathf.Clamp01(time / totalTime));
            yield return null;
        }
        transform.localScale = Vector3.one;
    }

    void Awake()
    {
        init();
    }


}
